@@ -14,16 +14,21 @@ GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
+import moderngl
 import numpy as np
+from moderngl_window.opengl.vao import VAO
 
 from aitviewer.scene.material import Material
 from aitviewer.scene.node import Node
-from aitviewer.shaders import get_depth_only_program, get_fragmap_program, get_outline_program, get_sphere_instanced_program
+from aitviewer.shaders import (
+    get_depth_only_program,
+    get_fragmap_program,
+    get_outline_program,
+    get_sphere_instanced_program,
+)
 from aitviewer.utils.decorators import hooked
 from aitviewer.utils.utils import set_lights_in_program, set_material_properties
 
-from moderngl_window.opengl.vao import VAO
-import moderngl
 
 def _create_sphere(radius=1.0, rings=16, sectors=32):
     """
@@ -54,12 +59,8 @@ def _create_sphere(radius=1.0, rings=16, sectors=32):
     i = 0
     for r in range(rings - 1):
         for s in range(sectors - 1):
-            faces[i] = np.array([r * sectors + s,
-                                 (r + 1) * sectors + (s + 1),
-                                 r * sectors + (s + 1)])
-            faces[i + 1] = np.array([r * sectors + s,
-                                    (r + 1) * sectors + s,
-                                    (r + 1) * sectors + (s + 1)])
+            faces[i] = np.array([r * sectors + s, (r + 1) * sectors + (s + 1), r * sectors + (s + 1)])
+            faces[i + 1] = np.array([r * sectors + s, (r + 1) * sectors + s, (r + 1) * sectors + (s + 1)])
             i += 2
 
     return vertices, faces
@@ -68,14 +69,17 @@ def _create_sphere(radius=1.0, rings=16, sectors=32):
 class Spheres(Node):
     """Render some simple spheres."""
 
-    def __init__(self,
-                 positions,
-                 radius=0.01,
-                 color=(0.0, 0.0, 1.0, 1.0),
-                 rings=16,
-                 sectors=32,
-                 icon="\u008d",
-                 **kwargs):
+    def __init__(
+        self,
+        positions,
+        radius=0.01,
+        color=(0.0, 0.0, 1.0, 1.0),
+        rings=16,
+        sectors=32,
+        icon="\u008d",
+        cast_shadow=False,
+        **kwargs,
+    ):
         """
         Initializer.
         :param positions: A numpy array of shape (F, N, 3) or (N, 3) containing N sphere positions for F time steps.
@@ -90,12 +94,15 @@ class Spheres(Node):
 
         # Define a default material in case there is None.
         if isinstance(color, tuple) or len(color.shape) == 1:
-            kwargs['material'] = kwargs.get('material', Material(color=color, ambient=0.2))
-            self.sphere_colors = kwargs['material'].color
+            kwargs["material"] = kwargs.get("material", Material(color=color, ambient=0.2))
+            self.sphere_colors = kwargs["material"].color
         else:
             assert color.shape[1] == 4 and positions.shape[1] == color.shape[0]
             self.sphere_colors = color
-        super().__init__(n_frames=positions.shape[0], icon=icon, **kwargs)
+
+        if "n_frames" not in kwargs:
+            kwargs["n_frames"] = positions.shape[0]
+        super().__init__(icon=icon, **kwargs)
 
         self._sphere_positions = positions
         self.radius = radius
@@ -111,6 +118,7 @@ class Spheres(Node):
         self.outline = True
         self.fragmap = True
         self.depth_prepass = True
+        self.cast_shadow = cast_shadow
 
     @property
     def bounds(self):
@@ -131,7 +139,7 @@ class Spheres(Node):
         if len(self._sphere_colors.shape) == 1:
             return np.full((self.n_spheres * self.n_vertices, 4), self._sphere_colors)
         else:
-            return np.tile(self._sphere_colors, (self.n_vertices,1))
+            return np.tile(self._sphere_colors, (self.n_vertices, 1))
 
     def color_one(self, index, color):
         new_colors = np.tile(np.array(self.material.color), (self.n_spheres, 1))
@@ -180,6 +188,7 @@ class Spheres(Node):
         if len(pos.shape) == 2:
             pos = pos[np.newaxis]
         self._sphere_positions = pos
+        self.n_frames = len(self._sphere_positions)
         self.redraw()
 
     def on_frame_update(self):
@@ -197,8 +206,8 @@ class Spheres(Node):
         self.depth_only_program = get_depth_only_program(vs_path)
         self.fragmap_program = get_fragmap_program(vs_path)
 
-        self.vbo_vertices = ctx.buffer(self.vertices.astype('f4').tobytes())
-        self.vbo_indices = ctx.buffer(self.faces.astype('i4').tobytes())
+        self.vbo_vertices = ctx.buffer(self.vertices.astype("f4").tobytes())
+        self.vbo_indices = ctx.buffer(self.faces.astype("i4").tobytes())
 
         self.vbo_instance_position = ctx.buffer(reserve=self.n_spheres * 12)
         self.vbo_instance_color = ctx.buffer(reserve=self.n_spheres * 16)
@@ -223,15 +232,21 @@ class Spheres(Node):
         prog = self.prog
         prog["radius"] = self.radius
         if len(self._sphere_colors.shape) == 1:
-            prog['use_uniform_color'] = True
-            prog['uniform_color'] = tuple(self._sphere_colors)
+            prog["use_uniform_color"] = True
+            prog["uniform_color"] = tuple(self._sphere_colors)
         else:
-            prog['use_uniform_color'] = False
-        prog['draw_edges'].value = 1.0 if self.draw_edges else 0.0
-        prog['win_size'].value = kwargs['window_size']
+            prog["use_uniform_color"] = False
+        prog["draw_edges"].value = 1.0 if self.draw_edges else 0.0
+        prog["win_size"].value = kwargs["window_size"]
+        prog["clip_control"].value = (0, 0, 0)
 
         self.set_camera_matrices(prog, camera, **kwargs)
-        set_lights_in_program(prog, kwargs['lights'], kwargs['shadows_enabled'], kwargs['ambient_strength'])
+        set_lights_in_program(
+            prog,
+            kwargs["lights"],
+            kwargs["shadows_enabled"],
+            kwargs["ambient_strength"],
+        )
         set_material_properties(prog, self.material)
         self.receive_shadow(prog, **kwargs)
         self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_spheres)
@@ -243,12 +258,23 @@ class Spheres(Node):
             self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_spheres)
 
     def gui(self, imgui):
-        _, self.radius = imgui.drag_float('Radius', self.radius, 0.01,
-                                        min_value=0.001,
-                                        max_value=10.0, format='%.3f')
+        _, self.radius = imgui.drag_float("Radius", self.radius, 0.01, min_value=0.001, max_value=10.0, format="%.3f")
         super().gui(imgui)
 
     @hooked
     def release(self):
         if self.is_renderable:
             self.vao.release()
+
+    def update_frames(self, positions, frames):
+        self.sphere_positions[frames] = positions
+        self.redraw()
+
+    def add_frames(self, positions):
+        if len(positions.shape) == 2:
+            positions = positions[np.newaxis]
+        self.sphere_positions = np.append(self.sphere_positions, positions, axis=0)
+
+    def remove_frames(self, frames):
+        self.sphere_positions = np.delete(self.sphere_positions, frames, axis=0)
+        self.redraw()
