@@ -14,21 +14,25 @@ GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
-import numpy as np
-import joblib
 import os
-
 from abc import ABC, abstractmethod
+
+import joblib
+import numpy as np
+from trimesh.transformations import rotation_matrix
+
 from aitviewer.configuration import CONFIG as C
 from aitviewer.renderables.lines import Lines
 from aitviewer.renderables.meshes import Meshes
 from aitviewer.renderables.rigid_bodies import RigidBodies
-from aitviewer.scene.camera_utils import look_at
-from aitviewer.scene.camera_utils import orthographic_projection
-from aitviewer.scene.camera_utils import perspective_projection
+from aitviewer.scene.camera_utils import (
+    look_at,
+    normalize,
+    orthographic_projection,
+    perspective_projection,
+)
 from aitviewer.scene.node import Node
 from aitviewer.utils.decorators import hooked
-from trimesh.transformations import rotation_matrix
 
 
 def _transform_vector(transform, vector):
@@ -36,6 +40,11 @@ def _transform_vector(transform, vector):
     return (transform @ np.concatenate([vector, np.array([1])]))[:3]
 
 
+def _transform_direction(transform, vector):
+    """Apply affine transformation (4-by-4 matrix) to a 3D directon."""
+    return (transform @ np.concatenate([vector, np.array([0])]))[:3]
+
+
 class CameraInterface(ABC):
     """
     An abstract class which describes the interface expected by the viewer for using this object as a camera
@@ -94,34 +103,42 @@ class Camera(Node, CameraInterface):
     A base camera object that provides rendering of a camera mesh and visualization of the camera frustum and coordinate
     system. Subclasses of this class must implement the CameraInterface abstract methods.
     """
-    def __init__(self, inactive_color=(0.5, 0.5, 0.5, 1), active_color=(0.6, 0.1, 0.1, 1), viewer=None, **kwargs):
-        """ Initializer
+
+    def __init__(
+        self,
+        inactive_color=(0.5, 0.5, 0.5, 1),
+        active_color=(0.6, 0.1, 0.1, 1),
+        viewer=None,
+        **kwargs,
+    ):
+        """Initializer
         :param inactive_color: Color that will be used for rendering this object when inactive
         :param active_color:   Color that will be used for rendering this object when active
         :param viewer: The current viewer, if not None the gui for this object will show a button for viewing from this
          camera in the viewer
         """
-        super(Camera, self).__init__(icon='\u0084', gui_material=False, **kwargs)
+        super(Camera, self).__init__(icon="\u0084", gui_material=False, **kwargs)
 
         # Camera object geometry
-        vertices = np.array([
-            # Body
-            [ 0,  0, 0],
-            [-1, -1, 1],
-            [-1,  1, 1],
-            [ 1, -1, 1],
-            [ 1,  1, 1],
-
-            # Triangle front
-            [ 0.5,  1.1, 1],
-            [-0.5,  1.1, 1],
-            [   0,    2, 1],
-
-            # Triangle back
-            [ 0.5,  1.1, 1],
-            [-0.5,  1.1, 1],
-            [   0,    2, 1],
-        ], dtype=np.float32)
+        vertices = np.array(
+            [
+                # Body
+                [0, 0, 0],
+                [-1, -1, 1],
+                [-1, 1, 1],
+                [1, -1, 1],
+                [1, 1, 1],
+                # Triangle front
+                [0.5, 1.1, 1],
+                [-0.5, 1.1, 1],
+                [0, 2, 1],
+                # Triangle back
+                [0.5, 1.1, 1],
+                [-0.5, 1.1, 1],
+                [0, 2, 1],
+            ],
+            dtype=np.float32,
+        )
 
         # Scale dimensions
         vertices[:, 0] *= 0.05
@@ -137,23 +154,31 @@ class Camera(Node, CameraInterface):
         # Reverse x too to maintain a consistent triangle winding
         vertices[:, 0] *= -1
 
-        faces = np.array([
-            [ 0, 1, 2],
-            [ 0, 2, 4],
-            [ 0, 4, 3],
-            [ 0, 3, 1],
-            [ 1, 3, 2],
-            [ 4, 2, 3],
-            [ 5, 6, 7],
-            [ 8, 10, 9],
-        ])
+        faces = np.array(
+            [
+                [0, 1, 2],
+                [0, 2, 4],
+                [0, 4, 3],
+                [0, 3, 1],
+                [1, 3, 2],
+                [4, 2, 3],
+                [5, 6, 7],
+                [8, 10, 9],
+            ]
+        )
 
         self._active = False
         self.active_color = active_color
         self.inactive_color = inactive_color
 
-        self.mesh = Meshes(vertices, faces, cast_shadow=False, flat_shading=True,
-                           rotation=kwargs.get('rotation'), is_selectable=False)
+        self.mesh = Meshes(
+            vertices,
+            faces,
+            cast_shadow=False,
+            flat_shading=True,
+            rotation=kwargs.get("rotation"),
+            is_selectable=False,
+        )
         self.mesh.color = self.inactive_color
         self.add(self.mesh, show_in_hierarchy=False)
 
@@ -229,32 +254,51 @@ class Camera(Node, CameraInterface):
 
             # Comput z coordinate of a point at the given distance
             view_p = np.array([0.0, 0.0, -distance])
-            ndc_p = (ndc_from_view @ np.concatenate([view_p, np.array([1])]))
+            ndc_p = ndc_from_view @ np.concatenate([view_p, np.array([1])])
 
             # Compute z after perspective division
             z = ndc_p[2] / ndc_p[3]
 
-            lines = np.array([
-                [-1, -1, -1], [-1,  1, -1],
-                [-1, -1,  z], [-1,  1,  z],
-                [ 1, -1, -1], [ 1,  1, -1],
-                [ 1, -1,  z], [ 1,  1,  z],
-
-                [-1, -1, -1], [-1, -1, z],
-                [-1,  1, -1], [-1,  1, z],
-                [ 1, -1, -1], [ 1, -1, z],
-                [ 1,  1, -1], [ 1,  1, z],
-
-                [-1, -1, -1], [ 1, -1, -1],
-                [-1, -1,  z], [ 1, -1,  z],
-                [-1,  1, -1], [ 1,  1, -1],
-                [-1,  1,  z], [ 1,  1,  z],
-            ], dtype=np.float32)
+            lines = np.array(
+                [
+                    [-1, -1, -1],
+                    [-1, 1, -1],
+                    [-1, -1, z],
+                    [-1, 1, z],
+                    [1, -1, -1],
+                    [1, 1, -1],
+                    [1, -1, z],
+                    [1, 1, z],
+                    [-1, -1, -1],
+                    [-1, -1, z],
+                    [-1, 1, -1],
+                    [-1, 1, z],
+                    [1, -1, -1],
+                    [1, -1, z],
+                    [1, 1, -1],
+                    [1, 1, z],
+                    [-1, -1, -1],
+                    [1, -1, -1],
+                    [-1, -1, z],
+                    [1, -1, z],
+                    [-1, 1, -1],
+                    [1, 1, -1],
+                    [-1, 1, z],
+                    [1, 1, z],
+                ],
+                dtype=np.float32,
+            )
 
             lines = np.apply_along_axis(transform, 1, lines)
             all_lines[i] = lines
 
-        self.frustum = Lines(all_lines, r_base=0.005, mode='lines', color=(0.1, 0.1, 0.1, 1), cast_shadow=False)
+        self.frustum = Lines(
+            all_lines,
+            r_base=0.005,
+            mode="lines",
+            color=(0.1, 0.1, 0.1, 1),
+            cast_shadow=False,
+        )
         self.add(self.frustum, show_in_hierarchy=False)
 
         ori = np.eye(3, dtype=np.float)
@@ -291,7 +335,13 @@ class Camera(Node, CameraInterface):
         path_spheres = RigidBodies(all_points, all_oris, radius=0.01, length=0.1, color=(0.92, 0.68, 0.2, 1.0))
         # Create lines only if there is more than one frame in the sequence.
         if self.n_frames > 1:
-            path_lines = Lines(all_points, color=(0, 0, 0, 1), r_base=0.003, mode='line_strip', cast_shadow=False)
+            path_lines = Lines(
+                all_points,
+                color=(0, 0, 0, 1),
+                r_base=0.003,
+                mode="line_strip",
+                cast_shadow=False,
+            )
         else:
             path_lines = None
 
@@ -310,17 +360,17 @@ class Camera(Node, CameraInterface):
         # the frustum and coordinate system visualization.
         self.mesh.render_outline(*args, **kwargs)
 
-    def view_from_camera(self):
+    def view_from_camera(self, viewport):
         """If the viewer is specified for this camera, change the current view to view from this camera"""
         if self.viewer:
             self.hide_path()
             self.hide_frustum()
-            self.viewer.set_temp_camera(self)
+            self.viewer.set_temp_camera(self, viewport)
 
     def gui(self, imgui):
         if self.viewer:
             if imgui.button("View from camera"):
-                self.view_from_camera()
+                self.view_from_camera(self.viewer.viewports[0])
 
         u, show = imgui.checkbox("Show path", self.path is not None)
         if u:
@@ -329,10 +379,10 @@ class Camera(Node, CameraInterface):
             else:
                 self.hide_path()
 
-    def gui_context_menu(self, imgui):
+    def gui_context_menu(self, imgui, x: int, y: int):
         if self.viewer:
             if imgui.menu_item("View from camera", shortcut=None, selected=False, enabled=True)[1]:
-                self.view_from_camera()
+                self.view_from_camera(self.viewer.get_viewport_at_position(x, y))
 
         u, show = imgui.checkbox("Show path", self.path is not None)
         if u:
@@ -348,8 +398,19 @@ class WeakPerspectiveCamera(Camera):
     The camera is positioned at (0,0,1) axis aligned and looks towards the negative z direction following the OpenGL
     conventions.
     """
-    def __init__(self, scale, translation, cols, rows, near=C.znear, far=C.zfar, viewer=None, **kwargs):
-        """ Initializer.
+
+    def __init__(
+        self,
+        scale,
+        translation,
+        cols,
+        rows,
+        near=None,
+        far=None,
+        viewer=None,
+        **kwargs,
+    ):
+        """Initializer.
         :param scale: A np array of scale parameters [sx, sy] of shape (2) or a sequence of parameters of shape (N, 2)
         :param translation: A np array of translation parameters [tx, ty] of shape (2) or a sequence of parameters of
           shape (N, 2).
@@ -361,7 +422,7 @@ class WeakPerspectiveCamera(Camera):
         :param far: Distance of the far plane from the camera.
         :param viewer: the current viewer, if not None the gui for this object will show a button for viewing from
           this camera in the viewer.
-         """
+        """
         if len(scale.shape) == 1:
             scale = scale[np.newaxis]
 
@@ -370,7 +431,7 @@ class WeakPerspectiveCamera(Camera):
 
         assert scale.shape[0] == translation.shape[0], "Number of frames in scale and translation must match"
 
-        kwargs['gui_affine'] = False
+        kwargs["gui_affine"] = False
         super(WeakPerspectiveCamera, self).__init__(n_frames=scale.shape[0], viewer=viewer, **kwargs)
 
         self.scale_factor = scale
@@ -378,13 +439,13 @@ class WeakPerspectiveCamera(Camera):
 
         self.cols = cols
         self.rows = rows
-        self.near = near
-        self.far = far
+        self.near = near if near is not None else C.znear
+        self.far = far if far is not None else C.zfar
         self.viewer = viewer
 
-        self.position =  np.array([0, 0, 1], dtype=np.float32)
-        self._right   =  np.array([1, 0, 0], dtype=np.float32)
-        self._up      =  np.array([0, 1, 0], dtype=np.float32)
+        self.position = np.array([0, 0, 1], dtype=np.float32)
+        self._right = np.array([1, 0, 0], dtype=np.float32)
+        self._up = np.array([0, 1, 0], dtype=np.float32)
         self._forward = -np.array([0, 0, 1], dtype=np.float32)
 
     @property
@@ -407,12 +468,14 @@ class WeakPerspectiveCamera(Camera):
         camera_ar = self.cols / self.rows
         ar = camera_ar / window_ar
 
-        P = np.array([
-            [sx * ar,  0, 0,  tx * sx * ar],
-            [      0, sy, 0,      -ty * sy],
-            [      0, 0, -1,             0],
-            [      0, 0,  0,             1],
-        ])
+        P = np.array(
+            [
+                [sx * ar, 0, 0, tx * sx * ar],
+                [0, sy, 0, -ty * sy],
+                [0, 0, -1, 0],
+                [0, 0, 0, 1],
+            ]
+        )
 
         znear, zfar = self.near, self.far
         P[2][2] = 2.0 / (znear - zfar)
@@ -421,9 +484,9 @@ class WeakPerspectiveCamera(Camera):
         V = look_at(self.position, self.forward, np.array([0, 1, 0]))
 
         # Update camera matrices
-        self.projection_matrix = P.astype('f4')
-        self.view_matrix = V.astype('f4')
-        self.view_projection_matrix = np.matmul(P, V).astype('f4')
+        self.projection_matrix = P.astype("f4")
+        self.view_matrix = V.astype("f4")
+        self.view_projection_matrix = np.matmul(P, V).astype("f4")
 
     @hooked
     def gui(self, imgui):
@@ -435,7 +498,7 @@ class WeakPerspectiveCamera(Camera):
                 self.hide_frustum()
 
     @hooked
-    def gui_context_menu(self, imgui):
+    def gui_context_menu(self, imgui, x: int, y: int):
         u, show = imgui.checkbox("Show frustum", self.frustum is not None)
         if u:
             if show:
@@ -446,14 +509,25 @@ class WeakPerspectiveCamera(Camera):
         imgui.spacing()
         imgui.separator()
         imgui.spacing()
-        super(Camera, self).gui_context_menu(imgui)
+        super(Camera, self).gui_context_menu(imgui, x, y)
 
 
 class OpenCVCamera(Camera):
-    """ A camera described by extrinsics and intrinsics in the format used by OpenCV """
+    """A camera described by extrinsics and intrinsics in the format used by OpenCV"""
 
-    def __init__(self, K, Rt, cols, rows, dist_coeffs=None, near=C.znear, far=C.zfar, viewer=None, **kwargs):
-        """ Initializer.
+    def __init__(
+        self,
+        K,
+        Rt,
+        cols,
+        rows,
+        dist_coeffs=None,
+        near=None,
+        far=None,
+        viewer=None,
+        **kwargs,
+    ):
+        """Initializer.
         :param K:  A np array of camera intrinsics in the format used by OpenCV (3, 3) or (N, 3, 3), one for each frame.
         :param Rt: A np array of camera extrinsics in the format used by OpenCV (3, 4) or (N, 3, 4), one for each frame.
         :param dist_coeffs: Lens distortion coefficients in the format used by OpenCV (5).
@@ -463,17 +537,18 @@ class OpenCVCamera(Camera):
         :param far: Distance of the far plane from the camera.
         :param viewer: The current viewer, if not None the gui for this object will show a button for viewing from this
           camera in the viewer.
-         """
+        """
         self.K = K if len(K.shape) == 3 else K[np.newaxis]
         self.Rt = Rt if len(Rt.shape) == 3 else Rt[np.newaxis]
 
         assert len(self.Rt.shape) == 3
         assert len(self.K.shape) == 3
 
-        assert self.K.shape[0] == 1 or self.Rt.shape[0] == 1 or self.K.shape[0] == self.Rt.shape[0], (
-            f"extrinsics and intrinsics array shape mismatch: {self.Rt.shape} and {self.K.shape}")
+        assert (
+            self.K.shape[0] == 1 or self.Rt.shape[0] == 1 or self.K.shape[0] == self.Rt.shape[0]
+        ), f"extrinsics and intrinsics array shape mismatch: {self.Rt.shape} and {self.K.shape}"
 
-        kwargs['gui_affine'] = False
+        kwargs["gui_affine"] = False
         super(OpenCVCamera, self).__init__(viewer=viewer, n_frames=max(self.K.shape[0], self.Rt.shape[0]), **kwargs)
         self.position = self.current_position
         self.rotation = self.current_rotation
@@ -482,8 +557,8 @@ class OpenCVCamera(Camera):
         self.cols = cols
         self.rows = rows
 
-        self.near = near
-        self.far = far
+        self.near = near if near is not None else C.znear
+        self.far = far if far is not None else C.zfar
 
     def on_frame_update(self):
         self.position = self.current_position
@@ -541,27 +616,31 @@ class OpenCVCamera(Camera):
         near, far = self.near, self.far
 
         # Compute number of columns that we would need in the image to preserve the aspect ratio
-        window_cols =  width / height * rows
+        window_cols = width / height * rows
 
         # Offset to center the image on the x direction
         x_offset = (window_cols - cols) * 0.5
 
         # Calibration matrix with added Z information and adapted to OpenGL coordinate
         # system which has (0,0) at center and Y pointing up
-        Kgl = np.array([
-            [-K[0,0],       0, -(cols - K[0, 2]) - x_offset,            0],
-            [      0, -K[1,1],             (rows - K[1, 2]),            0],
-            [      0,       0,                -(near + far),-(near * far)],
-            [      0,       0,                           -1,            0],
-        ])
+        Kgl = np.array(
+            [
+                [-K[0, 0], 0, -(cols - K[0, 2]) - x_offset, 0],
+                [0, -K[1, 1], (rows - K[1, 2]), 0],
+                [0, 0, -(near + far), -(near * far)],
+                [0, 0, -1, 0],
+            ]
+        )
 
         # Transformation from pixel coordinates to normalized device coordinates used by OpenGL
-        NDC = np.array([
-            [-2 / window_cols,         0,                 0,                             1],
-            [               0, -2 / rows,                 0,                            -1],
-            [               0,         0,  2 / (far - near),  -(far + near) / (far - near)],
-            [               0,         0,                 0,                             1],
-        ])
+        NDC = np.array(
+            [
+                [-2 / window_cols, 0, 0, 1],
+                [0, -2 / rows, 0, -1],
+                [0, 0, 2 / (far - near), -(far + near) / (far - near)],
+                [0, 0, 0, 1],
+            ]
+        )
 
         P = NDC @ Kgl
 
@@ -570,12 +649,12 @@ class OpenCVCamera(Camera):
     def update_matrices(self, width, height):
         V, P = self.compute_opengl_view_projection(width, height)
 
-        #Update camera matrices
-        self.projection_matrix = P.astype('f4')
-        self.view_matrix = V.astype('f4')
-        self.view_projection_matrix = np.matmul(P, V).astype('f4')
+        # Update camera matrices
+        self.projection_matrix = P.astype("f4")
+        self.view_matrix = V.astype("f4")
+        self.view_projection_matrix = np.matmul(P, V).astype("f4")
 
-    def to_pinhole_camera(self, target_distance=5, **kwargs) -> 'PinholeCamera':
+    def to_pinhole_camera(self, target_distance=5, **kwargs) -> "PinholeCamera":
         """
         Returns a PinholeCamera object with positions and targets computed from this camera.
         :param target_distance: distance from the camera at which the target of the PinholeCamera is placed.
@@ -606,8 +685,17 @@ class OpenCVCamera(Camera):
         # because the PiholeCamera does not support sequences of fov values.
         fov = np.rad2deg(2 * np.arctan(self.K[0, 1, 2] / self.K[0, 1, 1]))
 
-        return PinholeCamera(positions, targets, self.cols, self.rows, fov=fov,
-                             near=self.near, far=self.far, viewer=self.viewer, **kwargs)
+        return PinholeCamera(
+            positions,
+            targets,
+            self.cols,
+            self.rows,
+            fov=fov,
+            near=self.near,
+            far=self.far,
+            viewer=self.viewer,
+            **kwargs,
+        )
 
     @hooked
     def gui(self, imgui):
@@ -619,7 +707,7 @@ class OpenCVCamera(Camera):
                 self.hide_frustum()
 
     @hooked
-    def gui_context_menu(self, imgui):
+    def gui_context_menu(self, imgui, x: int, y: int):
         u, show = imgui.checkbox("Show frustum", self.frustum is not None)
         if u:
             if show:
@@ -630,18 +718,31 @@ class OpenCVCamera(Camera):
         imgui.spacing()
         imgui.separator()
         imgui.spacing()
-        super(Camera, self).gui_context_menu(imgui)
+        super(Camera, self).gui_context_menu(imgui, x, y)
 
 
 class PinholeCamera(Camera):
     """
     Your classic pinhole camera.
     """
-    def __init__(self, position, target, cols, rows, fov=45, near=C.znear, far=C.zfar, viewer=None, **kwargs):
+
+    def __init__(
+        self,
+        position,
+        target,
+        cols,
+        rows,
+        fov=45,
+        near=None,
+        far=None,
+        viewer=None,
+        **kwargs,
+    ):
         positions = position if len(position.shape) == 2 else position[np.newaxis]
         targets = target if len(target.shape) == 2 else target[np.newaxis]
-        assert positions.shape[0] == 1 or targets.shape[0] == 1 or positions.shape[0] == targets.shape[0], (
-            f"position and target array shape mismatch: {positions.shape} and {targets.shape}")
+        assert (
+            positions.shape[0] == 1 or targets.shape[0] == 1 or positions.shape[0] == targets.shape[0]
+        ), f"position and target array shape mismatch: {positions.shape} and {targets.shape}"
 
         self._world_up = np.array([0.0, 1.0, 0.0])
         self._targets = targets
@@ -650,8 +751,8 @@ class PinholeCamera(Camera):
         self.cols = cols
         self.rows = rows
 
-        self.near = near
-        self.far = far
+        self.near = near if near is not None else C.znear
+        self.far = far if far is not None else C.zfar
         self.fov = fov
 
     @property
@@ -678,7 +779,6 @@ class PinholeCamera(Camera):
     def rotation(self):
         return np.array([-self.right, self.up, -self.forward]).T
 
-
     def update_matrices(self, width, height):
         # Compute projection matrix.
         P = perspective_projection(np.deg2rad(self.fov), width / height, self.near, self.far)
@@ -687,9 +787,9 @@ class PinholeCamera(Camera):
         V = look_at(self.position, self.current_target, self._world_up)
 
         # Update camera matrices.
-        self.projection_matrix = P.astype('f4')
-        self.view_matrix = V.astype('f4')
-        self.view_projection_matrix = np.matmul(P, V).astype('f4')
+        self.projection_matrix = P.astype("f4")
+        self.view_matrix = V.astype("f4")
+        self.view_projection_matrix = np.matmul(P, V).astype("f4")
 
     def to_opencv_camera(self, **kwargs) -> OpenCVCamera:
         """
@@ -713,16 +813,36 @@ class PinholeCamera(Camera):
         Rts[:, 1:3, :] *= -1.0
 
         # Compute intrinsics.
-        f = 1. / np.tan(np.radians(self.fov / 2))
+        f = 1.0 / np.tan(np.radians(self.fov / 2))
         c0 = np.array([cols / 2.0, rows / 2.0])
-        K = np.array([[f * 0.5 * rows, 0., c0[0]], [0., f * 0.5 * rows, c0[1]], [0., 0., 1.]])
+        K = np.array(
+            [
+                [f * 0.5 * rows, 0.0, c0[0]],
+                [0.0, f * 0.5 * rows, c0[1]],
+                [0.0, 0.0, 1.0],
+            ]
+        )
 
-        return OpenCVCamera(K, Rts, cols, rows, near=self.near, far=self.far, viewer=self.viewer, **kwargs)
+        return OpenCVCamera(
+            K,
+            Rts,
+            cols,
+            rows,
+            near=self.near,
+            far=self.far,
+            viewer=self.viewer,
+            **kwargs,
+        )
 
     def gui_affine(self, imgui):
-        """ Render GUI for affine transformations"""
+        """Render GUI for affine transformations"""
         # Position controls
-        u, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 0.1, format='%.2f')
+        u, pos = imgui.drag_float3(
+            "Position##pos{}".format(self.unique_name),
+            *self.position,
+            0.1,
+            format="%.2f",
+        )
         if u:
             self.position = pos
 
@@ -736,7 +856,7 @@ class PinholeCamera(Camera):
                 self.hide_frustum()
 
     @hooked
-    def gui_context_menu(self, imgui):
+    def gui_context_menu(self, imgui, x: int, y: int):
         u, show = imgui.checkbox("Show frustum", self.frustum is not None)
         if u:
             if show:
@@ -747,7 +867,7 @@ class PinholeCamera(Camera):
         imgui.spacing()
         imgui.separator()
         imgui.spacing()
-        super(Camera, self).gui_context_menu(imgui)
+        super(Camera, self).gui_context_menu(imgui, x, y)
 
 
 class ViewerCamera(CameraInterface):
@@ -756,7 +876,7 @@ class ViewerCamera(CameraInterface):
     This camera also supports orbiting, panning and generating camera rays.
     """
 
-    def __init__(self, fov=45, orthographic=None, znear=C.znear, zfar=C.zfar):
+    def __init__(self, fov=45, orthographic=None, znear=None, zfar=None):
         super(ViewerCamera, self).__init__()
         self.fov = fov
         self.is_ortho = orthographic is not None
@@ -764,20 +884,30 @@ class ViewerCamera(CameraInterface):
 
         # Default camera settings.
         self._position = np.array([0.0, 0.0, 2.5])
-        self.target = np.array([0.0, 0.0, 0.0])
+        self._target = np.array([0.0, 0.0, 0.0])
         self._up = np.array([0.0, 1.0, 0.0])
 
         self.ZOOM_FACTOR = 4
         self.ROT_FACTOR = 0.0025
         self.PAN_FACTOR = 0.01
 
-        self.near = znear
-        self.far = zfar
+        self.near = znear if znear is not None else C.znear
+        self.far = zfar if zfar is not None else C.zfar
 
-        # GUI options
-        self.name = 'Camera'
-        self.icon = '\u0084'
+        # Controls options.
+        self.constant_speed = 1.0
+        self._control_modes = ["turntable", "trackball", "first_person"]
+        self._control_mode = "turntable"
+        self._trackball_start_view_inverse = None
+        self._trackball_start_hit = None
+        self._trackball_start_position = None
+        self._trackball_start_up = None
 
+        # GUI options.
+        self.name = "Camera"
+        self.icon = "\u0084"
+
+        # Animation.
         self.animating = False
         self._animation_t = 0.0
         self._animation_time = 0.0
@@ -786,6 +916,28 @@ class ViewerCamera(CameraInterface):
         self._animation_start_target = None
         self._animation_end_target = None
 
+    @property
+    def control_mode(self):
+        return self._control_mode
+
+    @control_mode.setter
+    def control_mode(self, mode):
+        if mode not in self._control_modes:
+            raise ValueError(f"Invalid camera mode: {mode}")
+        if mode == "first_person" or mode == "turntable":
+            self.up = (0, 1, 0)
+        self._control_mode = mode
+
+    def copy(self):
+        camera = ViewerCamera(self.fov, self.ortho_size, self.near, self.far)
+        camera.is_ortho = self.is_ortho
+        camera.position = self.position
+        camera.target = self.target
+        camera.up = self.up
+        camera.constant_speed = self.constant_speed
+        camera.control_mode = self.control_mode
+        return camera
+
     @property
     def position(self):
         return self._position
@@ -796,8 +948,7 @@ class ViewerCamera(CameraInterface):
 
     @property
     def forward(self):
-        forward = self.target - self.position
-        return forward / np.linalg.norm(forward)
+        return normalize(self.target - self.position)
 
     @property
     def up(self):
@@ -805,46 +956,53 @@ class ViewerCamera(CameraInterface):
 
     @up.setter
     def up(self, up):
-        self._up = up
+        self._up = np.array(up, dtype=np.float32).copy()
 
     @property
     def right(self):
-        right = np.cross(self.up, self.forward)
-        return right / np.linalg.norm(right)
+        return normalize(np.cross(self.up, self.forward))
+
+    @property
+    def target(self):
+        return self._target
+
+    @target.setter
+    def target(self, t):
+        self._target = np.array(t, np.float32).copy()
 
     def save_cam(self):
         """Saves the current camera parameters"""
-        cam_dir = C.export_dir + '/camera_params/'
+        cam_dir = C.export_dir + "/camera_params/"
         if not os.path.exists(cam_dir):
             os.makedirs(cam_dir)
 
         cam_dict = {}
-        cam_dict['position'] = self.position
-        cam_dict['target'] = self.target
-        cam_dict['up'] = self.up
-        cam_dict['ZOOM_FACTOR'] = self.ZOOM_FACTOR
-        cam_dict['ROT_FACTOR'] = self.ROT_FACTOR
-        cam_dict['PAN_FACTOR'] = self.PAN_FACTOR
-        cam_dict['near'] = self.near
-        cam_dict['far'] = self.far
+        cam_dict["position"] = self.position
+        cam_dict["target"] = self.target
+        cam_dict["up"] = self.up
+        cam_dict["ZOOM_FACTOR"] = self.ZOOM_FACTOR
+        cam_dict["ROT_FACTOR"] = self.ROT_FACTOR
+        cam_dict["PAN_FACTOR"] = self.PAN_FACTOR
+        cam_dict["near"] = self.near
+        cam_dict["far"] = self.far
 
-        joblib.dump(cam_dict, cam_dir+'cam_params.pkl')
+        joblib.dump(cam_dict, cam_dir + "cam_params.pkl")
 
     def load_cam(self):
         """Loads the camera parameters"""
-        cam_dir = C.export_dir + '/camera_params/'
+        cam_dir = C.export_dir + "/camera_params/"
         if not os.path.exists(cam_dir):
-            print('camera config does not exist')
+            print("camera config does not exist")
         else:
-            cam_dict = joblib.load(cam_dir + 'cam_params.pkl')
-            self.position = cam_dict['position']
-            self.target = cam_dict['target']
-            self.up = cam_dict['up']
-            self.ZOOM_FACTOR = cam_dict['ZOOM_FACTOR']
-            self.ROT_FACTOR = cam_dict['ROT_FACTOR']
-            self.PAN_FACTOR = cam_dict['PAN_FACTOR']
-            self.near = cam_dict['near']
-            self.far = cam_dict['far']
+            cam_dict = joblib.load(cam_dir + "cam_params.pkl")
+            self.position = cam_dict["position"]
+            self.target = cam_dict["target"]
+            self.up = cam_dict["up"]
+            self.ZOOM_FACTOR = cam_dict["ZOOM_FACTOR"]
+            self.ROT_FACTOR = cam_dict["ROT_FACTOR"]
+            self.PAN_FACTOR = cam_dict["PAN_FACTOR"]
+            self.near = cam_dict["near"]
+            self.far = cam_dict["far"]
 
     def update_matrices(self, width, height):
         # Compute projection matrix.
@@ -859,11 +1017,11 @@ class ViewerCamera(CameraInterface):
         V = look_at(self.position, self.target, self.up)
 
         # Update camera matrices.
-        self.projection_matrix = P.astype('f4')
-        self.view_matrix = V.astype('f4')
-        self.view_projection_matrix = np.matmul(P, V).astype('f4')
+        self.projection_matrix = P.astype("f4")
+        self.view_matrix = V.astype("f4")
+        self.view_projection_matrix = np.matmul(P, V).astype("f4")
 
-    def dolly_zoom(self, speed, move_target=False):
+    def dolly_zoom(self, speed, move_target=False, constant_speed=False):
         """
         Zoom by moving the camera along its view direction.
         If move_target is true the camera target will also move rigidly with the camera.
@@ -880,7 +1038,9 @@ class ViewerCamera(CameraInterface):
         # Adjust speed according to config
         speed *= C.camera_zoom_speed
 
-        if move_target:
+        if move_target or constant_speed:
+            if constant_speed:
+                norm = self.constant_speed * 20
             self.position += fwd * speed * norm
             self.target += fwd * speed * norm
         else:
@@ -893,7 +1053,7 @@ class ViewerCamera(CameraInterface):
 
     def pan(self, mouse_dx, mouse_dy):
         """Move the camera in the image plane."""
-        sideways = np.cross(self.forward, self.up)
+        sideways = normalize(np.cross(self.forward, self.up))
         up = np.cross(sideways, self.forward)
 
         # scale speed according to distance from target
@@ -908,8 +1068,16 @@ class ViewerCamera(CameraInterface):
         self.position += up * speed_y
         self.target += up * speed_y
 
-    def rotate_azimuth_elevation(self, mouse_dx, mouse_dy):
-        """Rotate the camera left-right and up-down (roll is not allowed)."""
+    def rotate_azimuth(self, angle):
+        """Rotate around camera's up-axis by given angle (in radians)."""
+        if np.abs(angle) < 1e-8:
+            return
+        cam_pose = np.linalg.inv(self.view_matrix)
+        y_axis = cam_pose[:3, 1]
+        rot = rotation_matrix(angle, y_axis, self.target)
+        self.position = _transform_vector(rot, self.position)
+
+    def _rotation_from_mouse_delta(self, mouse_dx: int, mouse_dy: int):
         z_axis = -self.forward
         dot = np.dot(z_axis, self.up)
         rot = np.eye(4)
@@ -925,25 +1093,92 @@ class ViewerCamera(CameraInterface):
         y_axis = np.cross(self.forward, self.right)
         x_speed = self.ROT_FACTOR / 10 if 1 - np.abs(dot) < 0.01 else self.ROT_FACTOR
         rot = rotation_matrix(x_speed * -mouse_dx, y_axis, self.target) @ rot
+        return rot
 
+    def rotate_azimuth_elevation(self, mouse_dx: int, mouse_dy: int):
+        """Rotate the camera position left-right and up-down orbiting around the target (roll is not allowed)."""
+        rot = self._rotation_from_mouse_delta(mouse_dx, mouse_dy)
         self.position = _transform_vector(rot, self.position)
 
-    def rotate_azimuth(self, angle):
-        """Rotate around camera's up-axis by given angle (in radians)."""
-        if np.abs(angle) < 1e-8:
+    def rotate_first_person(self, mouse_dx: int, mouse_dy: int):
+        """Rotate the camera target left-right and up-down (roll is not allowed)."""
+        rot = self._rotation_from_mouse_delta(mouse_dx, mouse_dy)
+        self.target = _transform_direction(rot, self.target - self.position) + self.position
+
+    def intersect_trackball(self, x: int, y: int, width: int, height: int):
+        """
+        Return intersection of a line passing through the mouse position at pixel coordinates x, y
+        and the trackball as a point in world coordinates.
+        """
+        # Transform mouse coordinates from -1 to 1
+        nx = 2 * (x + 0.5) / width - 1
+        ny = 1 - 2 * (y + 0.5) / height
+
+        # Adjust coordinates for the larger side of the viewport rectangle.
+        if width > height:
+            nx *= width / height
+        else:
+            ny *= height / width
+
+        s = nx * nx + ny * ny
+        if s <= 0.5:
+            # Sphere intersection
+            nz = np.sqrt(1 - s)
+        else:
+            # Hyperboloid intersection.
+            nz = 1 / (2 * np.sqrt(s))
+
+        # Return intersection position in world coordinates.
+        return self._trackball_start_view_inverse @ np.array((nx, ny, nz))
+
+    def rotate_trackball(self, x: int, y: int, width: int, height: int):
+        """Rotate the camera with trackball controls. Must be called after rotate_start()."""
+        # Compute points on trackball.
+        start = self._trackball_start_hit
+        current = self.intersect_trackball(x, y, width, height)
+        dist = np.linalg.norm(current - start)
+
+        # Skip if starting and current point are too close.
+        if dist < 1e-6:
             return
-        cam_pose = np.linalg.inv(self.view_matrix)
-        y_axis = cam_pose[:3, 1]
-        rot = rotation_matrix(angle, y_axis, self.target)
-        self.position = _transform_vector(rot, self.position)
+
+        # Compute axis of rotation as the vector perpendicular to the plane spanned by the
+        # vectors connecting the origin to the two points.
+        axis = normalize(np.cross(current, start))
+
+        # Compute angle as the angle between the two vectors, if they are too far away we use the distance
+        # between them instead, this makes it continue to rotate when dragging the mouse further away.
+        angle = max(np.arccos(np.dot(normalize(current), normalize(start))), dist)
+
+        # Compute resulting rotation and apply it to the starting position and up vector.
+        rot = rotation_matrix(angle, axis, self.target)
+        self.position = _transform_vector(rot, self._trackball_start_position)
+        self.up = _transform_direction(rot, self._trackball_start_up)
+
+    def rotate_start(self, x: int, y: int, width: int, height: int):
+        """Start rotating the camera. Called on mouse button press."""
+        if self.control_mode == "trackball":
+            self._trackball_start_view_inverse = look_at(self.position, self.target, self.up)[:3, :3].T
+            self._trackball_start_hit = self.intersect_trackball(x, y, width, height)
+            self._trackball_start_position = self.position
+            self._trackball_start_up = self.up
+
+    def rotate(self, x: int, y: int, mouse_dx: int, mouse_dy: int, width: int, height: int):
+        """Rotate the camera. Called on mouse movement."""
+        if self.control_mode == "turntable":
+            self.rotate_azimuth_elevation(mouse_dx, mouse_dy)
+        elif self.control_mode == "trackball":
+            self.rotate_trackball(x, y, width, height)
+        elif self.control_mode == "first_person":
+            self.rotate_first_person(mouse_dx, mouse_dy)
 
     def get_ray(self, x, y, width, height):
         """Construct a ray going through the middle of the given pixel."""
         w, h = width, height
 
         # Pixel in (-1, 1) range.
-        screen_x = (2 * (x + 0.5) / w - 1)
-        screen_y = (1 - 2 * (y + 0.5) / h)
+        screen_x = 2 * (x + 0.5) / w - 1
+        screen_y = 1 - 2 * (y + 0.5) / h
 
         # Scale to actual image plane size.
         scale = self.ortho_size if self.is_ortho else np.tan(np.deg2rad(self.fov) / 2)
@@ -990,9 +1225,15 @@ class ViewerCamera(CameraInterface):
             self.target = self._animation_start_target * (1 - t) + self._animation_end_target * t
 
     def gui(self, imgui):
-        _, self.is_ortho = imgui.checkbox('Orthographic Camera', self.is_ortho)
-        _, self.fov = imgui.slider_float('Camera FOV##fov', self.fov, 0.1, 180.0, '%.1f')
-        _, position = imgui.drag_float3('Position', *self.position)
-        _, target = imgui.drag_float3('Target', *self.target)
-        self.position = np.array(position)
-        self.target = np.array(target)
+        _, self.is_ortho = imgui.checkbox("Orthographic Camera", self.is_ortho)
+        _, self.fov = imgui.slider_float("Camera FOV##fov", self.fov, 0.1, 180.0, "%.1f")
+        _, self.position = imgui.drag_float3("Position", *self.position)
+        _, self.target = imgui.drag_float3("Target", *self.target)
+        _, self.up = imgui.drag_float3("Up", *self.up)
+
+        imgui.spacing()
+        # Note: must be kept in sync with self._control_modes.
+        control_modes_labels = ["Turntable", "Trackball", "First Person"]
+        u, idx = imgui.combo("Control mode", self._control_modes.index(self.control_mode), control_modes_labels)
+        if u and idx >= 0 and idx <= len(self._control_modes):
+            self.control_mode = self._control_modes[idx]
