@@ -14,31 +14,34 @@ GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
+from functools import lru_cache
+
 import moderngl
 import numpy as np
 
 from aitviewer.configuration import CONFIG as C
 from aitviewer.scene.material import Material
 from aitviewer.utils.so3 import euler2rot_numpy, rot2euler_numpy
-from functools import lru_cache
 
 
 class Node(object):
     """Interface for nodes."""
 
-    def __init__(self,
-                 name=None,
-                 icon=None,
-                 position=None,
-                 rotation=None,
-                 scale=1.0,
-                 color=(0.5, 0.5, 0.5, 1.0),
-                 material=None,
-                 is_selectable=True,
-                 gui_affine=True,
-                 gui_material=True,
-                 enabled_frames=None,
-                 n_frames=1):
+    def __init__(
+        self,
+        name=None,
+        icon=None,
+        position=None,
+        rotation=None,
+        scale=1.0,
+        color=(0.5, 0.5, 0.5, 1.0),
+        material=None,
+        is_selectable=True,
+        gui_affine=True,
+        gui_material=True,
+        enabled_frames=None,
+        n_frames=1,
+    ):
         """
         :param name: Name of the node
         :param icon: Custom Node Icon using custom Icon font
@@ -67,19 +70,26 @@ class Node(object):
         n_scales = self._scales.shape[0]
 
         if n_frames > 1:
-            assert n_positions == 1 or n_frames == n_positions, (f"Number of position frames"
-                f" ({n_positions}) must be 1 or match number of Node frames {n_frames}")
-            assert n_rotations == 1 or n_frames == n_rotations, (f"Number of rotations frames"
-                f" ({n_rotations}) must be 1 or match number of Node frames {n_frames}")
-            assert n_scales == 1 or n_frames == n_scales, (f"Number of scales frames"
-                f" ({n_scales}) must be 1 or match number of Node frames {n_frames}")
+            assert n_positions == 1 or n_frames == n_positions, (
+                f"Number of position frames" f" ({n_positions}) must be 1 or match number of Node frames {n_frames}"
+            )
+            assert n_rotations == 1 or n_frames == n_rotations, (
+                f"Number of rotations frames" f" ({n_rotations}) must be 1 or match number of Node frames {n_frames}"
+            )
+            assert n_scales == 1 or n_frames == n_scales, (
+                f"Number of scales frames" f" ({n_scales}) must be 1 or match number of Node frames {n_frames}"
+            )
         else:
             n_frames = max(n_positions, n_rotations, n_scales)
-            assert ((n_positions == 1 or n_positions == n_frames) and
-                    (n_rotations == 1 or n_rotations == n_frames) and
-                    (n_scales == 1 or n_scales == n_frames)), (f"Number of position"
-                        f"({n_positions}), rotation ({n_rotations}) and scale ({n_scales})"
-                        "frames must be 1 or match.")
+            assert (
+                (n_positions == 1 or n_positions == n_frames)
+                and (n_rotations == 1 or n_rotations == n_frames)
+                and (n_scales == 1 or n_scales == n_frames)
+            ), (
+                f"Number of position"
+                f"({n_positions}), rotation ({n_rotations}) and scale ({n_scales})"
+                "frames must be 1 or match."
+            )
 
         # Frames
         self._n_frames = n_frames
@@ -87,8 +97,10 @@ class Node(object):
         self.model_matrix = self.get_local_transform()
         self._enabled_frames = enabled_frames
         if self._enabled_frames is not None:
-            assert np.count_nonzero(self._enabled_frames) == n_frames, (f"Number of non-zero elements in enabled_frames"
-                f" ({np.count_nonzero(self._enabled_frames)}) must match number of frames in sequence ({n_frames})")
+            assert np.count_nonzero(self._enabled_frames) == n_frames, (
+                f"Number of non-zero elements in enabled_frames"
+                f" ({np.count_nonzero(self._enabled_frames)}) must match number of frames in sequence ({n_frames})"
+            )
             # Create an array that maps from the true frame id (counting also disabled frames) to the index of the
             # first existing frame in the sequence.
             self._enabled_frame_id = np.cumsum(self._enabled_frames) - 1
@@ -113,27 +125,41 @@ class Node(object):
         self.outline = False
 
         # Programs for render passes. Subclasses are responsible for setting these.
-        self.depth_only_program = None # Required for depth_prepass and cast_shadow passes
-        self.fragmap_program = None    # Required for framap pass
-        self.outline_program = None    # Required for outline pass
+        self.depth_only_program = None  # Required for depth_prepass and cast_shadow passes
+        self.fragmap_program = None  # Required for fragmap pass
+        self.outline_program = None  # Required for outline pass
 
         # GUI
         self.name = name if name is not None else type(self).__name__
         self.uid = C.next_gui_id()
         self.unique_name = self.name + "{}".format(self.uid)
-        self.icon = icon if icon is not None else '\u0082'
+        self.icon = icon if icon is not None else "\u0082"
         self._enabled = True
         self._expanded = False
         self.gui_controls = {
-            'affine': {'fn': self.gui_affine, 'icon': '\u009b', 'is_visible': gui_affine},
-            'material': {'fn': self.gui_material, 'icon': '\u0088', 'is_visible': gui_material},
-            'animation': {'fn': self.gui_animation, 'icon': '\u0098', 'is_visible': (lambda: self._n_frames > 1)()},
-            'io': {'fn': self.gui_io, 'icon': '\u009a', 'is_visible': (lambda: self.gui_io.__func__ is not Node.gui_io)()}
+            "affine": {
+                "fn": self.gui_affine,
+                "icon": "\u009b",
+                "is_visible": gui_affine,
+            },
+            "material": {
+                "fn": self.gui_material,
+                "icon": "\u0088",
+                "is_visible": gui_material,
+            },
+            "animation": {
+                "fn": self.gui_animation,
+                "icon": "\u0098",
+                "is_visible": (lambda: self._n_frames > 1)(),
+            },
+            "io": {
+                "fn": self.gui_io,
+                "icon": "\u009a",
+                "is_visible": (lambda: self.gui_io.__func__ is not Node.gui_io)(),
+            },
         }
-        self.gui_modes = {
-            'view': {'title': ' View', 'fn': self.gui_mode_view, 'icon': '\u0099'}
-        }
-        self._selected_mode = 'view'
+        self.gui_modes = {"view": {"title": " View", "fn": self.gui_mode_view, "icon": "\u0099"}}
+        self._selected_mode = "view"
         self._show_in_hierarchy = True
         self.is_selectable = is_selectable
 
@@ -221,21 +247,18 @@ class Node(object):
 
         scale = np.diag([scale, scale, scale, 1])
 
-        return (trans @ rotation @ scale).astype('f4')
+        return (trans @ rotation @ scale).astype("f4")
 
     def get_local_transform(self):
         """Construct local transform as a 4x4 matrix from this node's position, orientation and scale."""
-        return self._compute_transform(
-               tuple(self.position),
-               tuple(map(tuple, self.rotation)),
-               self.scale)
+        return self._compute_transform(tuple(self.position), tuple(map(tuple, self.rotation)), self.scale)
 
     def update_transform(self, parent_transform=None):
         """Update the model matrix of this node and all of its descendants."""
         if parent_transform is None:
             self.model_matrix = self.get_local_transform()
         else:
-            self.model_matrix = parent_transform.astype('f4') @ self.get_local_transform()
+            self.model_matrix = parent_transform.astype("f4") @ self.get_local_transform()
 
         for n in self.nodes:
             n.update_transform(self.model_matrix)
@@ -250,7 +273,7 @@ class Node(object):
 
     @property
     def bounds(self):
-        """ The bounds in the format ((x_min, x_max), (y_min, y_max), (z_min, z_max)) """
+        """The bounds in the format ((x_min, x_max), (y_min, y_max), (z_min, z_max))"""
         return np.array([[0, 0], [0, 0], [0, 0]])
 
     @property
@@ -271,10 +294,13 @@ class Node(object):
         assert len(points.shape) == 3
 
         # Compute min and max coordinates of the bounding box ignoring NaNs.
-        val = np.array([
-            [np.nanmin(points[:, :, 0]), np.nanmax(points[:, :, 0])],
-            [np.nanmin(points[:, :, 1]), np.nanmax(points[:, :, 1])],
-            [np.nanmin(points[:, :, 2]), np.nanmax(points[:, :, 2])]])
+        val = np.array(
+            [
+                [np.nanmin(points[:, :, 0]), np.nanmax(points[:, :, 0])],
+                [np.nanmin(points[:, :, 1]), np.nanmax(points[:, :, 1])],
+                [np.nanmin(points[:, :, 2]), np.nanmax(points[:, :, 2])],
+            ]
+        )
 
         # If any of the elements is NaN return an empty bounding box.
         if np.isnan(val).any():
@@ -352,7 +378,7 @@ class Node(object):
 
     def on_before_frame_update(self):
         """Called when the current frame is about to change, 'self.current_frame_id' still has the id of the
-           previous frame."""
+        previous frame."""
         pass
 
     def on_frame_update(self):
@@ -362,11 +388,7 @@ class Node(object):
     def add(self, *nodes, **kwargs):
         self._add_nodes(*nodes, **kwargs)
 
-    def _add_node(self,
-                  n: 'Node',
-                  show_in_hierarchy=True,
-                  expanded=False,
-                  enabled=True):
+    def _add_node(self, n: "Node", show_in_hierarchy=True, expanded=False, enabled=True):
         """
         Add a single node
         :param show_in_hierarchy: Whether to show the node in the scene hierarchy.
@@ -431,69 +453,105 @@ class Node(object):
         pass
 
     def gui_modes(self, imgui):
-        """ Render GUI with toolbar (tools) for this particular node"""
+        """Render GUI with toolbar (tools) for this particular node"""
 
     def gui_animation(self, imgui):
-        """ Render GUI for animation related settings"""
+        """Render GUI for animation related settings"""
 
         if self._enabled_frames is None:
             if self.n_frames > 1:
-                u, fid = imgui.slider_int('Frame##r_{}'.format(self.unique_name),
-                                        self.current_frame_id, min_value=0, max_value=self.n_frames - 1)
+                u, fid = imgui.slider_int(
+                    "Frame##r_{}".format(self.unique_name),
+                    self.current_frame_id,
+                    min_value=0,
+                    max_value=self.n_frames - 1,
+                )
                 if u:
                     self.current_frame_id = fid
         else:
-            u, fid = imgui.slider_int('Frame##r_{}'.format(self.unique_name),
-                                    self._internal_frame_id, min_value=0, max_value=self._enabled_frames.shape[0] - 1)
+            u, fid = imgui.slider_int(
+                "Frame##r_{}".format(self.unique_name),
+                self._internal_frame_id,
+                min_value=0,
+                max_value=self._enabled_frames.shape[0] - 1,
+            )
             if u:
                 self.current_frame_id = fid
 
     def gui_affine(self, imgui):
-        """ Render GUI for affine transformations"""
+        """Render GUI for affine transformations"""
         # Position controls
-        up, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 1e-2, format='%.2f')
+        up, pos = imgui.drag_float3(
+            "Position##pos{}".format(self.unique_name),
+            *self.position,
+            1e-2,
+            format="%.2f",
+        )
         if up:
             self.position = pos
 
         # Rotation controls
         euler_angles = rot2euler_numpy(self.rotation[np.newaxis], degrees=True)[0]
-        ur, euler_angles = imgui.drag_float3('Rotation##pos{}'.format(self.unique_name), *euler_angles, 1e-2, format='%.2f')
+        ur, euler_angles = imgui.drag_float3(
+            "Rotation##pos{}".format(self.unique_name),
+            *euler_angles,
+            1e-2,
+            format="%.2f",
+        )
         if ur:
             self.rotation = euler2rot_numpy(np.array(euler_angles)[np.newaxis], degrees=True)[0]
 
         # Scale controls
-        us, scale = imgui.drag_float('Scale##scale{}'.format(self.unique_name), self.scale, 1e-2, min_value=0.001,
-                                    max_value=100.0, format='%.3f')
+        us, scale = imgui.drag_float(
+            "Scale##scale{}".format(self.unique_name),
+            self.scale,
+            1e-2,
+            min_value=0.001,
+            max_value=100.0,
+            format="%.3f",
+        )
         if us:
             self.scale = scale
 
     def gui_material(self, imgui):
-        """ Render GUI with material properties """
+        """Render GUI with material properties"""
 
         # Color Control
-        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.material.color, show_alpha=True)
+        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.material.color)
         if uc:
             self.color = color
 
         # Diffuse
-        ud, diffuse = imgui.slider_float('Diffuse##diffuse{}'.format(self.unique_name), self.material.diffuse,  0.0, 1.0, '%.2f')
+        ud, diffuse = imgui.slider_float(
+            "Diffuse##diffuse{}".format(self.unique_name),
+            self.material.diffuse,
+            0.0,
+            1.0,
+            "%.2f",
+        )
         if ud:
             self.material.diffuse = diffuse
 
         # Ambient
-        ua, ambient = imgui.slider_float('Ambient##ambient{}'.format(self.unique_name), self.material.ambient,  0.0, 1.0, '%.2f')
+        ua, ambient = imgui.slider_float(
+            "Ambient##ambient{}".format(self.unique_name),
+            self.material.ambient,
+            0.0,
+            1.0,
+            "%.2f",
+        )
         if ua:
             self.material.ambient = ambient
 
     def gui_io(self, imgui):
-        """ Render GUI for import/export """
+        """Render GUI for import/export"""
         pass
 
     def gui_mode_view(self, imgui):
-        """ Render custom GUI for view mode """
+        """Render custom GUI for view mode"""
         pass
 
-    def gui_context_menu(self, imgui):
+    def gui_context_menu(self, imgui, x: int, y: int):
         _, self.enabled = imgui.checkbox("Enabled", self.enabled)
         if any([n._show_in_hierarchy for n in self.nodes]):
             imgui.spacing()
@@ -503,7 +561,7 @@ class Node(object):
                 if not n._show_in_hierarchy:
                     continue
                 if imgui.begin_menu(f"{n.name}##{n.uid}"):
-                    n.gui_context_menu(imgui)
+                    n.gui_context_menu(imgui, x, y)
                     imgui.end_menu()
 
     # Renderable
@@ -515,6 +573,7 @@ class Node(object):
             else:
                 func(self, *args, **kwargs)
                 self.is_renderable = True
+
         return _decorator
 
     def make_renderable(self, ctx):
@@ -535,15 +594,15 @@ class Node(object):
         pass
 
     def redraw(self, **kwargs):
-        """ Perform update and redraw operations. Push to the GPU when finished. Recursively redraw child nodes"""
+        """Perform update and redraw operations. Push to the GPU when finished. Recursively redraw child nodes"""
         for n in self.nodes:
             n.redraw(**kwargs)
 
     def set_camera_matrices(self, prog, camera, **kwargs):
         """Set the model view projection matrix in the given program."""
         # Transpose because np is row-major but OpenGL expects column-major.
-        prog['model_matrix'].write(self.model_matrix.T.astype('f4').tobytes())
-        prog['view_projection_matrix'].write(camera.get_view_projection_matrix().T.astype('f4').tobytes())
+        prog["model_matrix"].write(self.model_matrix.T.astype("f4").tobytes())
+        prog["view_projection_matrix"].write(camera.get_view_projection_matrix().T.astype("f4").tobytes())
 
     def receive_shadow(self, program, **kwargs):
         """
@@ -551,20 +610,20 @@ class Node(object):
         :param program: The shader program that can shade with shadows.
         :param kwargs: The render kwargs.
         """
-        if kwargs.get('shadows_enabled', False):
-            lights = kwargs['lights']
+        if kwargs.get("shadows_enabled", False):
+            lights = kwargs["lights"]
 
             for i, light in enumerate(lights):
                 if light.shadow_enabled and light.shadow_map:
                     light_matrix = light.mvp() @ self.model_matrix
-                    program[f'dirLights[{i}].matrix'].write(light_matrix.T.tobytes())
+                    program[f"dirLights[{i}].matrix"].write(light_matrix.T.tobytes())
 
                     # Bind shadowmap to slot i + 1, we reserve slot 0 for the mesh texture
                     # and use slots 1 to (#lights + 1) for shadow maps
                     light.shadow_map.use(location=i + 1)
 
             # Set sampler uniforms
-            uniform = program[f'shadow_maps']
+            uniform = program[f"shadow_maps"]
             uniform.value = 1 if uniform.array_length == 1 else [*range(1, len(lights) + 1)]
 
     def render_shadowmap(self, light_matrix):
@@ -572,8 +631,8 @@ class Node(object):
             return
 
         prog = self.depth_only_program
-        prog['model_matrix'].write(self.model_matrix.T.tobytes())
-        prog['view_projection_matrix'].write(light_matrix.T.tobytes())
+        prog["model_matrix"].write(self.model_matrix.T.tobytes())
+        prog["view_projection_matrix"].write(light_matrix.T.tobytes())
 
         self.render_positions(prog)
 
@@ -586,7 +645,7 @@ class Node(object):
         self.set_camera_matrices(prog, camera)
 
         # Render with the specified object uid, if None use the node uid instead.
-        prog['obj_id'] = uid or self.uid
+        prog["obj_id"] = uid or self.uid
 
         if self.backface_culling or self.backface_fragmap:
             ctx.enable(moderngl.CULL_FACE)
@@ -595,13 +654,13 @@ class Node(object):
 
         # If backface_fragmap is enabled for this node only render backfaces
         if self.backface_fragmap:
-            ctx.cull_face = 'front'
+            ctx.cull_face = "front"
 
         self.render_positions(prog)
 
         # Restore cull face to back
         if self.backface_fragmap:
-            ctx.cull_face = 'back'
+            ctx.cull_face = "back"
 
     def render_depth_prepass(self, camera, **kwargs):
         if not self.depth_prepass or self.depth_only_program is None:
@@ -613,7 +672,7 @@ class Node(object):
 
     def render_outline(self, ctx, camera):
         if self.outline and self.outline_program is not None:
-            prog  = self.outline_program
+            prog = self.outline_program
             self.set_camera_matrices(prog, camera)
 
             if self.backface_culling:
@@ -651,3 +710,12 @@ class Node(object):
         Handle shortcut key presses (if you are the selected object)
         """
         pass
+
+    def update_frames(self, *args, **kwargs):
+        pass
+
+    def add_frames(self, *args, **kwargs):
+        pass
+
+    def remove_frames(self, *args, **kwargs):
+        pass
