@@ -14,14 +14,15 @@ GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
-import numpy as np
-import torch
 import os
 import subprocess
 
+import numpy as np
+import torch
+from scipy.interpolate import CubicSpline
+
 from aitviewer.utils.so3 import aa2rot_torch as aa2rot
 from aitviewer.utils.so3 import rot2aa_torch as rot2aa
-from scipy.interpolate import CubicSpline
 
 
 def to_torch(x, dtype, device):
@@ -39,13 +40,14 @@ def to_numpy(x):
 
 
 def get_video_paths(video_path):
-    is_mp4 = video_path.endswith('.mp4')
-    is_gif = video_path.endswith('.gif')
-    if not (is_mp4 or is_gif):
-        video_path += '.mp4'
+    is_webm = video_path.endswith(".webm")
+    is_mp4 = video_path.endswith(".mp4")
+    is_gif = video_path.endswith(".gif")
+    if not (is_mp4 or is_gif or is_webm):
+        video_path += ".mp4"
         is_mp4 = True
 
-    suffix = '.gif' if is_gif else '.mp4'
+    suffix = os.path.splitext(video_path)[1]
 
     dir_of_file = os.path.dirname(os.path.abspath(video_path))
     if not os.path.exists(dir_of_file):
@@ -53,34 +55,44 @@ def get_video_paths(video_path):
 
     # Make sure we don't override an existing video.
     counter = 0
-    video_path_candidate = video_path.replace(suffix, f'_{counter}{suffix}')
+    video_path_candidate = video_path.replace(suffix, f"_{counter}{suffix}")
     while os.path.exists(video_path_candidate):
         counter += 1
-        video_path_candidate = video_path.replace(suffix, f'_{counter}{suffix}')
+        video_path_candidate = video_path.replace(suffix, f"_{counter}{suffix}")
 
-    if is_mp4:
+    if is_mp4 or is_webm:
         return video_path_candidate, None, is_gif
     else:
-        video_path_mp4 = video_path_candidate.replace(suffix, f'_temp.mp4')
-        return video_path_mp4, video_path_candidate, is_gif
+        video_path_temp = video_path_candidate.replace(suffix, f"_temp.mp4")
+        return video_path_temp, video_path_candidate, is_gif
 
 
 def video_to_gif(path_mp4, path_gif, remove=False):
-    command = ['ffmpeg',
-               '-i', path_mp4,
-               '-y',
-               '-filter_complex', f"[0:v] split [a][b];[a] palettegen=stats_mode=single [p];[b][p] paletteuse=new=1:dither=none",
-               path_gif]
+    command = [
+        "ffmpeg",
+        "-i",
+        path_mp4,
+        "-y",
+        "-vf",
+        "split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse",
+        path_gif,
+    ]
 
-    with open(os.devnull, 'w') as FNULL:
+    with open(os.devnull, "w") as FNULL:
         subprocess.Popen(command, stdout=FNULL, stderr=FNULL).wait()
 
     if remove:
         os.remove(path_mp4)
 
 
-def images_to_video(frame_dir, video_path, frame_format='frame_%06d.png', input_fps=60, output_fps=60,
-                    start_frame=0):
+def images_to_video(
+    frame_dir,
+    video_path,
+    frame_format="frame_%06d.png",
+    input_fps=60,
+    output_fps=60,
+    start_frame=0,
+):
     """Convert the rendered images into a video. The video path format determines whether this will be rendered as
     a GIF or an MP4 (default)."""
     if not os.path.exists(frame_dir):
@@ -91,21 +103,33 @@ def images_to_video(frame_dir, video_path, frame_format='frame_%06d.png', input_
     print("Rendering to video {}".format(os.path.abspath(path_mp4)))
 
     # Create mp4 from frames.
-    command = ['ffmpeg',
-               '-framerate', str(input_fps),  # must be this early in the command, otherwise it is not applied.
-               '-start_number', str(start_frame),
-               '-i', os.path.join(frame_dir, frame_format),
-               '-c:v', 'libx264',
-               '-preset', 'slow',
-               '-profile:v', 'high',
-               '-level:v', '4.0',
-               '-pix_fmt', 'yuv420p',
-               '-vf', 'pad=ceil(iw/2)*2:ceil(ih/2)*2',  # Avoid error when image res is not divisible by 2.
-               '-r', str(output_fps),
-               '-y',
-               path_mp4]
+    command = [
+        "ffmpeg",
+        "-framerate",
+        str(input_fps),  # must be this early in the command, otherwise it is not applied.
+        "-start_number",
+        str(start_frame),
+        "-i",
+        os.path.join(frame_dir, frame_format),
+        "-c:v",
+        "libx264",
+        "-preset",
+        "slow",
+        "-profile:v",
+        "high",
+        "-level:v",
+        "4.0",
+        "-pix_fmt",
+        "yuv420p",
+        "-vf",
+        "pad=ceil(iw/2)*2:ceil(ih/2)*2",  # Avoid error when image res is not divisible by 2.
+        "-r",
+        str(output_fps),
+        "-y",
+        path_mp4,
+    ]
 
-    with open(os.devnull, 'w') as FNULL:
+    with open(os.devnull, "w") as FNULL:
         subprocess.Popen(command, stdout=FNULL, stderr=FNULL).wait()
 
     # If GIF we first create a temporary mp4 and then convert that to GIF to reduce palette artifacts.
@@ -210,7 +234,7 @@ def compute_vertex_and_face_normals_sparse(vertices, faces, vertex_faces_sparse,
     """
     triangles = vertices[:, faces]
     vectors = np.diff(triangles, axis=2)
-    crosses = np.cross(vectors[:,:, 0], vectors[:,:, 1])
+    crosses = np.cross(vectors[:, :, 0], vectors[:, :, 1])
 
     normals = np.moveaxis(crosses, 0, 2)
     normals = normals.reshape((normals.shape[0], -1))
@@ -241,6 +265,7 @@ def spherical_coordinates_from_direction(v, degrees=False):
     else:
         return theta, phi
 
+
 def direction_from_spherical_coordinates(theta, phi, degrees=False):
     """
     Converts from spherical coordinates to a direction vector.
@@ -268,16 +293,16 @@ def direction_from_spherical_coordinates(theta, phi, degrees=False):
 def set_lights_in_program(prog, lights, shadows_enabled, ambient_strength):
     """Set program lighting from scene lights"""
     for i, light in enumerate(lights):
-        prog[f'dirLights[{i}].direction'].value = tuple(light.direction)
-        prog[f'dirLights[{i}].color'].value = light.light_color
-        prog[f'dirLights[{i}].strength'].value = light.strength if light.enabled else 0.0
-        prog[f'dirLights[{i}].shadow_enabled'].value = shadows_enabled and light.shadow_enabled
-    prog['ambient_strength'] = ambient_strength
+        prog[f"dirLights[{i}].direction"].value = tuple(light.direction)
+        prog[f"dirLights[{i}].color"].value = light.light_color
+        prog[f"dirLights[{i}].strength"].value = light.strength if light.enabled else 0.0
+        prog[f"dirLights[{i}].shadow_enabled"].value = shadows_enabled and light.shadow_enabled
+    prog["ambient_strength"] = ambient_strength
 
 
 def set_material_properties(prog, material):
-    prog['diffuse_coeff'].value = material.diffuse
-    prog['ambient_coeff'].value = material.ambient
+    prog["diffuse_coeff"].value = material.diffuse
+    prog["ambient_coeff"].value = material.ambient
 
 
 def compute_union_of_bounds(nodes):
@@ -304,7 +329,7 @@ def compute_union_of_current_bounds(nodes):
     return bounds
 
 
-def local_to_global(poses, parents, output_format='aa', input_format='aa'):
+def local_to_global(poses, parents, output_format="aa", input_format="aa"):
     """
     Convert relative joint angles to global ones by unrolling the kinematic chain.
     :param poses: A tensor of shape (N, N_JOINTS*3) defining the relative poses in angle-axis format.
@@ -313,11 +338,11 @@ def local_to_global(poses, parents, output_format='aa', input_format='aa'):
     :param input_format: 'aa' or 'rotmat'
     :return: The global joint angles as a tensor of shape (N, N_JOINTS*DOF).
     """
-    assert output_format in ['aa', 'rotmat']
-    assert input_format in ['aa', 'rotmat']
-    dof = 3 if input_format == 'aa' else 9
+    assert output_format in ["aa", "rotmat"]
+    assert input_format in ["aa", "rotmat"]
+    dof = 3 if input_format == "aa" else 9
     n_joints = poses.shape[-1] // dof
-    if input_format == 'aa':
+    if input_format == "aa":
         local_oris = aa2rot(poses.reshape((-1, 3)))
     else:
         local_oris = poses
@@ -333,7 +358,7 @@ def local_to_global(poses, parents, output_format='aa', input_format='aa'):
             local_rot = local_oris[..., j, :, :]
             global_oris[..., j, :, :] = torch.matmul(parent_rot, local_rot)
 
-    if output_format == 'aa':
+    if output_format == "aa":
         global_oris = rot2aa(global_oris.reshape((-1, 3, 3)))
         res = global_oris.reshape((-1, n_joints * 3))
     else:
